import types

from edsnlp.data.converters import set_current_tokenizer


def apply_basic_pipes(docs, pipes):
    for name, pipe, kwargs, tok in pipes:
        with set_current_tokenizer(tok):
            if hasattr(pipe, "batch_process"):
                docs = pipe.batch_process(docs)
            else:
                results = []
                for doc in docs:
                    res = pipe(doc, **kwargs)
                    if isinstance(res, types.GeneratorType):
                        results.extend(res)
                    else:
                        results.append(res)
                docs = results
    return docs



import edsnlp
from edsnlp.utils.span_getters import make_span_context_getter


def test_span_context_getter_symmetric(lang):
    nlp = edsnlp.blank(lang)
    nlp.add_pipe("eds.normalizer")
    nlp.add_pipe("eds.sentences")
    nlp.add_pipe("eds.matcher", config={"terms": {"sentence": "sentence"}})
    doc = nlp(
        "This is a sentence. "
        "This is another sentence. "
        "This is a third one. "
        "Last sentence."
    )

    span_getter = make_span_context_getter(
        context_words=2,
    )
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence. This",
        "This is another sentence. This",
        ". Last sentence.",
    ]

    span_getter = make_span_context_getter(
        context_words=0,
        context_sents=1,
    )
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence.",
        "This is another sentence.",
        "Last sentence.",
    ]

    span_getter = make_span_context_getter(
        context_words=0,
        context_sents=2,
    )
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence. This is another sentence.",
        "This is a sentence. This is another sentence. This is a third one.",
        "This is a third one. Last sentence.",
    ]


def test_span_context_getter_asymmetric(lang):
    nlp = edsnlp.blank(lang)
    nlp.add_pipe("eds.normalizer")
    nlp.add_pipe("eds.sentences")
    nlp.add_pipe("eds.matcher", config={"terms": {"animal": "kangaroo"}})
    doc = nlp(
        "This is a sentence. "
        "This is another sentence with a kangaroo. "
        "This is a third one. "
        "Last sentence."
    )

    span_getter = make_span_context_getter(context_words=2, context_sents=0)
    assert [span_getter(s).text for s in doc.ents] == [
        "with a kangaroo. This",
    ]

    span_getter = make_span_context_getter(context_words=(2, 1), context_sents=0)
    assert [span_getter(s).text for s in doc.ents] == [
        "with a kangaroo.",
    ]

    span_getter = make_span_context_getter(context_words=(1, 2), context_sents=0)
    assert [span_getter(s).text for s in doc.ents] == [
        "a kangaroo. This",
    ]

    span_getter = make_span_context_getter(context_words=0, context_sents=(1, 2))
    assert [span_getter(s).text for s in doc.ents] == [
        "This is another sentence with a kangaroo. This is a third one.",
    ]

    span_getter = make_span_context_getter(context_words=0, context_sents=(2, 2))
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence. This is another sentence with a kangaroo. This is a third one."  # noqa: E501
    ]

    span_getter = make_span_context_getter(context_words=0, context_sents=(1, 1))
    assert [span_getter(s).text for s in doc.ents] == [
        "This is another sentence with a kangaroo."
    ]

    span_getter = make_span_context_getter(context_words=(1000, 0), context_sents=0)
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence. This is another sentence with a kangaroo"
    ]

    span_getter = make_span_context_getter(
        context_words=(1000, 0), context_sents=(1, 2)
    )
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence. This is another sentence with a kangaroo. This is a third one."  # noqa: E501
    ]



"""`eds.tobacco` pipeline"""

from typing import Any, Dict, List, Optional, Union

from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanSetterArg
from edsnlp.utils.numbers import parse_digit

from ..alcohol.alcohol import AlcoholMatcher
from .patterns import default_patterns


class TobaccoMatcher(AlcoholMatcher):
    """
    The `eds.tobacco` pipeline component extracts mentions of tobacco consumption.

    ??? info "Details of the used patterns"
        ```{ .python .no-check }
        # fmt: off
        --8<-- "edsnlp/pipes/ner/behaviors/tobacco/patterns.py"
        # fmt: on
        ```

    Extensions
    ----------
    On each span `span` that match, the following attributes are available:

    - `span._.detailed_status`: either None or `"ABSTINENCE"`
    if the patient stopped its consumption
    - `span._.assigned`: dictionary with the following keys, if relevant:
        - `PA`: the mentioned *year-pack* (= *paquet-année*)
        - `secondhand`: if secondhand smoking
    - `span._.negation`: set to True when either
        - A pack-year value of 0 is extracted
        - A mention such as "tabac: 0" is found
        - The patient experiences secondhand smoking

    !!! warning "Use qualifiers !"
        Although the tobacco pipe sometime sets value for the `negation` attribute,
        *generic* qualifier should still be used after the pipe.


    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(
        eds.normalizer(
            accents=True,
            lowercase=True,
            quotes=True,
            spaces=True,
            pollution=dict(
                information=True,
                bars=True,
                biology=True,
                doctors=True,
                web=True,
                coding=True,
                footer=True,
            ),
        ),
    )
    nlp.add_pipe(eds.tobacco())
    ```

    Below are a few examples:

    --8<-- "docs/assets/fragments/tobacco-examples.md"


    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline object
    name : Optional[str]
        The name of the component
    patterns: Union[Dict[str, Any], List[Dict[str, Any]]]
        The patterns to use for matching
    label : str
        The label to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    Authors and citation
    --------------------
    The `eds.tobacco` component was developed by AP-HP's Data Science team with a
    team of medical experts, following the insights of the algorithm proposed
    by [@petitjean_2024].
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: str = "tobacco",
        *,
        patterns: Union[Dict[str, Any], List[Dict[str, Any]]] = default_patterns,
        label: str = "tobacco",
        span_setter: SpanSetterArg = {"ents": True, "tobacco": True},
    ):
        super().__init__(
            nlp=nlp,
            name=name,
            patterns=patterns,
            label=label,
            span_setter=span_setter,
        )

    def process(self, doc: Doc) -> List[Span]:
        for span in super().process(doc):
            if "secondhand" in span._.assigned.keys():
                span._.negation = True

            elif "PA" in span._.assigned.keys():
                pa = parse_digit(
                    span._.assigned["PA"],
                    atttr="NORM",
                    ignore_excluded=True,
                )
                if (pa == 0) and ("stopped" not in span._.assigned.keys()):
                    span._.negation = True

            yield span



number_terms = {
    "0.125": ["⅛"],
    "0.16666666": ["⅙"],
    "0.2": ["⅕"],
    "0.25": ["¼"],
    "0.3333333": ["⅓"],
    "0.5": ["½"],
    "2.5": ["21/2"],
    "1": ["un", "une"],
    "2": ["deux"],
    "3": ["trois"],
    "4": ["quatre"],
    "5": ["cinq"],
    "6": ["six"],
    "7": ["sept"],
    "8": ["huit"],
    "9": ["neuf"],
    "10": ["dix"],
    "11": ["onze"],
    "12": ["douze"],
    "13": ["treize"],
    "14": ["quatorze"],
    "15": ["quinze"],
    "16": ["seize"],
    "17": ["dix-sept", "dix sept"],
    "18": ["dix-huit", "dix huit"],
    "19": ["dix-neuf", "dix neuf"],
    "20": ["vingt", "vingts"],
    "30": ["trente"],
    "40": ["quarante"],
    "50": ["cinquante"],
    "60": ["soixante"],
    "70": ["soixante dix", "soixante-dix"],
    "80": ["quatre vingt", "quatre-vingt", "quatre vingts", "quatre-vingts"],
    "90": ["quatre vingt dix", "quatre-vingt-dix"],
    "100": ["cent"],
    "500": ["cinq cent", "cinq-cent"],
    "1000": ["mille", "milles"],
}


number_regex = r"""(?x)
# no digit or floating point number prefix before
(?<![0-9][.,]?)
# integer part like 123 or 1 234
(?:
    0
    |[1-9][0-9]*(?:\ \d{3})*
)
(?:
    # floating point surounded by spaces
    \ +[,.]\ +\d+
    # floating point w/o space
    | [,.]\d+
    # fractions
    | (?:\ /\ |/)[1-9][0-9]*(?:\ \d{3})*
)?"""


common_quantities = {
    "weight": {
        "unit": "kg",
        "unitless_patterns": [
            {
                "terms": ["poids", "poid", "pese", "pesant", "pesait", "pesent"],
                "ranges": [
                    {"min": 0, "max": 200, "unit": "kg"},
                    {"min": 200, "unit": "g"},
                ],
            }
        ],
    },
    "size": {
        "unit": "m",
        "unitless_patterns": [
            {
                "terms": [
                    "mesure",
                    "taille",
                    "mesurant",
                    "mesurent",
                    "mesurait",
                    "mesuree",
                    "hauteur",
                    "largeur",
                    "longueur",
                ],
                "ranges": [
                    {"min": 0, "max": 3, "unit": "m"},
                    {"min": 3, "unit": "cm"},
                ],
            }
        ],
    },
    "bmi": {
        "unit": "kg_per_m2",
        "unitless_patterns": [
            {"terms": ["imc", "bmi"], "ranges": [{"unit": "kg_per_m2"}]}
        ],
    },
    "volume": {"unit": "m3", "unitless_patterns": []},
}

unit_divisors = ["/", "par"]

stopwords = ["par", "sur", "de", "a", ",", "et", "-", "à"]

# Should we only make accented patterns and expect the user to use
# `eds.normalizer` component first ?
range_patterns = [
    ("De", "à"),
    ("De", "a"),
    ("de", "à"),
    ("de", "a"),
    ("Entre", "et"),
    ("entre", "et"),
    (None, "a"),
    (None, "à"),
    (None, "-"),
]

units_config = {
    "per_kg": {
        "dim": "mass",
        "degree": -1,
        "scale": 0.001,
        "terms": [
            "kg-1",
            "kilog-1",
            "kilo-g-1",
            "kgr-1",
            "kilogr-1",
            "kilo-gr-1",
            "kgr⁻¹",
            "kilogr⁻¹",
            "kilo-gr⁻¹",
            "kg⁻¹",
            "kilog⁻¹",
            "kilo-g⁻¹",
        ],
        "followed_by": None,
    },
    "per_hg": {
        "dim": "mass",
        "degree": -1,
        "scale": 0.01,
        "terms": [
            "hg-1",
            "hectog-1",
            "hecto-g-1",
            "hgr-1",
            "hectogr-1",
            "hecto-gr-1",
            "hgr⁻¹",
            "hectogr⁻¹",
            "hecto-gr⁻¹",
            "hg⁻¹",
            "hectog⁻¹",
            "hecto-g⁻¹",
        ],
        "followed_by": None,
    },
    "per_dag": {
        "dim": "mass",
        "degree": -1,
        "scale": 0.1,
        "terms": [
            "dag-1",
            "decag-1",
            "deca-g-1",
            "dagr-1",
            "decagr-1",
            "deca-gr-1",
            "dagr⁻¹",
            "decagr⁻¹",
            "deca-gr⁻¹",
            "dag⁻¹",
            "decag⁻¹",
            "deca-g⁻¹",
        ],
        "followed_by": None,
    },
    "per_dg": {
        "dim": "mass",
        "degree": -1,
        "scale": 10.0,
        "terms": [
            "dg-1",
            "decig-1",
            "deci-g-1",
            "dgr-1",
            "decigr-1",
            "deci-gr-1",
            "dgr⁻¹",
            "decigr⁻¹",
            "deci-gr⁻¹",
            "dg⁻¹",
            "decig⁻¹",
            "deci-g⁻¹",
        ],
        "followed_by": None,
    },
    "per_cg": {
        "dim": "mass",
        "degree": -1,
        "scale": 100.0,
        "terms": [
            "cg-1",
            "centig-1",
            "centi-g-1",
            "cgr-1",
            "centigr-1",
            "centi-gr-1",
            "cgr⁻¹",
            "centigr⁻¹",
            "centi-gr⁻¹",
            "cg⁻¹",
            "centig⁻¹",
            "centi-g⁻¹",
        ],
        "followed_by": None,
    },
    "per_mg": {
        "dim": "mass",
        "degree": -1,
        "scale": 1000.0,
        "terms": [
            "mg-1",
            "millig-1",
            "milli-g-1",
            "milig-1",
            "mili-g-1",
            "mgr-1",
            "milligr-1",
            "milli-gr-1",
            "miligr-1",
            "mili-gr-1",
            "mgr⁻¹",
            "milligr⁻¹",
            "milli-gr⁻¹",
            "miligr⁻¹",
            "mili-gr⁻¹",
            "mg⁻¹",
            "millig⁻¹",
            "milli-g⁻¹",
            "milig⁻¹",
            "mili-g⁻¹",
        ],
        "followed_by": None,
    },
    "per_μg": {
        "dim": "mass",
        "degree": -1,
        "scale": 1000000.0,
        "terms": [
            "μg-1",
            "ug-1",
            "µg-1",
            "microg-1",
            "micro-g-1",
            "μgr-1",
            "ugr-1",
            "µgr-1",
            "microgr-1",
            "micro-gr-1",
            "μgr⁻¹",
            "ugr⁻¹",
            "µgr⁻¹",
            "microgr⁻¹",
            "micro-gr⁻¹",
            "μg⁻¹",
            "ug⁻¹",
            "µg⁻¹",
            "microg⁻¹",
            "micro-g⁻¹",
        ],
        "followed_by": None,
    },
    "per_ng": {
        "dim": "mass",
        "degree": -1,
        "scale": 999999999.9999999,
        "terms": [
            "ng-1",
            "nanog-1",
            "nano-g-1",
            "ngr-1",
            "nanogr-1",
            "nano-gr-1",
            "ngr⁻¹",
            "nanogr⁻¹",
            "nano-gr⁻¹",
            "ng⁻¹",
            "nanog⁻¹",
            "nano-g⁻¹",
        ],
        "followed_by": None,
    },
    "per_pg": {
        "dim": "mass",
        "degree": -1,
        "scale": 1000000000000.0,
        "terms": [
            "pg-1",
            "picog-1",
            "pico-g-1",
            "pgr-1",
            "picogr-1",
            "pico-gr-1",
            "pgr⁻¹",
            "picogr⁻¹",
            "pico-gr⁻¹",
            "pg⁻¹",
            "picog⁻¹",
            "pico-g⁻¹",
        ],
        "followed_by": None,
    },
    "per_fg": {
        "dim": "mass",
        "degree": -1,
        "scale": 999999999999999.9,
        "terms": [
            "fg-1",
            "femtog-1",
            "femto-g-1",
            "fgr-1",
            "femtogr-1",
            "femto-gr-1",
            "fgr⁻¹",
            "femtogr⁻¹",
            "femto-gr⁻¹",
            "fg⁻¹",
            "femtog⁻¹",
            "femto-g⁻¹",
        ],
        "followed_by": None,
    },
    "per_g": {
        "dim": "mass",
        "degree": -1,
        "scale": 1.0,
        "terms": ["g-1", "gr-1", "gr⁻¹", "g⁻¹"],
        "followed_by": None,
    },
    "kg": {
        "dim": "mass",
        "degree": 1,
        "scale": 1000.0,
        "terms": [
            "kgramme",
            "kilogramme",
            "kilo-gramme",
            "kgrammes",
            "kilogrammes",
            "kilo-grammes",
            "kgr",
            "kilogr",
            "kilo-gr",
            "kg",
            "kilog",
            "kilo-g",
        ],
        "followed_by": "g",
    },
    "hg": {
        "dim": "mass",
        "degree": 1,
        "scale": 100.0,
        "terms": [
            "hgramme",
            "hectogramme",
            "hecto-gramme",
            "hgrammes",
            "hectogrammes",
            "hecto-grammes",
            "hgr",
            "hectogr",
            "hecto-gr",
            "hg",
            "hectog",
            "hecto-g",
        ],
        "followed_by": None,
    },
    "dag": {
        "dim": "mass",
        "degree": 1,
        "scale": 10.0,
        "terms": [
            "dagramme",
            "decagramme",
            "deca-gramme",
            "dagrammes",
            "decagrammes",
            "deca-grammes",
            "dagr",
            "decagr",
            "deca-gr",
            "dag",
            "decag",
            "deca-g",
        ],
        "followed_by": None,
    },
    "dg": {
        "dim": "mass",
        "degree": 1,
        "scale": 0.1,
        "terms": [
            "dgramme",
            "decigramme",
            "deci-gramme",
            "dgrammes",
            "decigrammes",
            "deci-grammes",
            "dgr",
            "decigr",
            "deci-gr",
            "dg",
            "decig",
            "deci-g",
        ],
        "followed_by": None,
    },
    "cg": {
        "dim": "mass",
        "degree": 1,
        "scale": 0.01,
        "terms": [
            "cgramme",
            "centigramme",
            "centi-gramme",
            "cgrammes",
            "centigrammes",
            "centi-grammes",
            "cgr",
            "centigr",
            "centi-gr",
            "cg",
            "centig",
            "centi-g",
        ],
        "followed_by": None,
    },
    "mg": {
        "dim": "mass",
        "degree": 1,
        "scale": 0.001,
        "terms": [
            "mgramme",
            "milligramme",
            "milli-gramme",
            "miligramme",
            "mili-gramme",
            "mgrammes",
            "milligrammes",
            "milli-grammes",
            "miligrammes",
            "mili-grammes",
            "mgr",
            "milligr",
            "milli-gr",
            "miligr",
            "mili-gr",
            "mg",
            "millig",
            "milli-g",
            "milig",
            "mili-g",
        ],
        "followed_by": None,
    },
    "μg": {
        "dim": "mass",
        "degree": 1,
        "scale": 1e-06,
        "terms": [
            "μgramme",
            "ugramme",
            "µgramme",
            "microgramme",
            "micro-gramme",
            "μgrammes",
            "ugrammes",
            "µgrammes",
            "microgrammes",
            "micro-grammes",
            "μgr",
            "ugr",
            "µgr",
            "microgr",
            "micro-gr",
            "μg",
            "ug",
            "µg",
            "microg",
            "micro-g",
        ],
        "followed_by": None,
    },
    "ng": {
        "dim": "mass",
        "degree": 1,
        "scale": 1e-09,
        "terms": [
            "ngramme",
            "nanogramme",
            "nano-gramme",
            "ngrammes",
            "nanogrammes",
            "nano-grammes",
            "ngr",
            "nanogr",
            "nano-gr",
            "ng",
            "nanog",
            "nano-g",
        ],
        "followed_by": None,
    },
    "pg": {
        "dim": "mass",
        "degree": 1,
        "scale": 1e-12,
        "terms": [
            "pgramme",
            "picogramme",
            "pico-gramme",
            "pgrammes",
            "picogrammes",
            "pico-grammes",
            "pgr",
            "picogr",
            "pico-gr",
            "pg",
            "picog",
            "pico-g",
        ],
        "followed_by": None,
    },
    "fg": {
        "dim": "mass",
        "degree": 1,
        "scale": 1e-15,
        "terms": [
            "fgramme",
            "femtogramme",
            "femto-gramme",
            "fgrammes",
            "femtogrammes",
            "femto-grammes",
            "fgr",
            "femtogr",
            "femto-gr",
            "fg",
            "femtog",
            "femto-g",
        ],
        "followed_by": None,
    },
    "g": {
        "dim": "mass",
        "degree": 1,
        "scale": 1,
        "terms": ["gramme", "grammes", "gr", "g"],
        "followed_by": None,
    },
    "per_km": {
        "dim": "length",
        "degree": -1,
        "scale": 0.001,
        "terms": ["km-1", "kilom-1", "kilo-m-1", "km⁻¹", "kilom⁻¹", "kilo-m⁻¹"],
        "followed_by": None,
    },
    "per_hm": {
        "dim": "length",
        "degree": -1,
        "scale": 0.01,
        "terms": ["hm-1", "hectom-1", "hecto-m-1", "hm⁻¹", "hectom⁻¹", "hecto-m⁻¹"],
        "followed_by": None,
    },
    "per_dam": {
        "dim": "length",
        "degree": -1,
        "scale": 0.1,
        "terms": ["dam-1", "decam-1", "deca-m-1", "dam⁻¹", "decam⁻¹", "deca-m⁻¹"],
        "followed_by": None,
    },
    "per_dm": {
        "dim": "length",
        "degree": -1,
        "scale": 10.0,
        "terms": ["dm-1", "decim-1", "deci-m-1", "dm⁻¹", "decim⁻¹", "deci-m⁻¹"],
        "followed_by": None,
    },
    "per_cm": {
        "dim": "length",
        "degree": -1,
        "scale": 100.0,
        "terms": ["cm-1", "centim-1", "centi-m-1", "cm⁻¹", "centim⁻¹", "centi-m⁻¹"],
        "followed_by": None,
    },
    "per_mm": {
        "dim": "length",
        "degree": -1,
        "scale": 1000.0,
        "terms": [
            "mm-1",
            "millim-1",
            "milli-m-1",
            "milim-1",
            "mili-m-1",
            "mm⁻¹",
            "millim⁻¹",
            "milli-m⁻¹",
            "milim⁻¹",
            "mili-m⁻¹",
        ],
        "followed_by": None,
    },
    "per_μm": {
        "dim": "length",
        "degree": -1,
        "scale": 1000000.0,
        "terms": [
            "μm-1",
            "um-1",
            "µm-1",
            "microm-1",
            "micro-m-1",
            "μm⁻¹",
            "um⁻¹",
            "µm⁻¹",
            "microm⁻¹",
            "micro-m⁻¹",
        ],
        "followed_by": None,
    },
    "per_nm": {
        "dim": "length",
        "degree": -1,
        "scale": 1e9,
        "terms": ["nm-1", "nanom-1", "nano-m-1", "nm⁻¹", "nanom⁻¹", "nano-m⁻¹"],
        "followed_by": None,
    },
    "per_pm": {
        "dim": "length",
        "degree": -1,
        "scale": 1e12,
        "terms": ["pm-1", "picom-1", "pico-m-1", "pm⁻¹", "picom⁻¹", "pico-m⁻¹"],
        "followed_by": None,
    },
    "per_fm": {
        "dim": "length",
        "degree": -1,
        "scale": 1e15,
        "terms": ["fm-1", "femtom-1", "femto-m-1", "fm⁻¹", "femtom⁻¹", "femto-m⁻¹"],
        "followed_by": None,
    },
    "per_m": {
        "dim": "length",
        "degree": -1,
        "scale": 1.0,
        "terms": ["m-1", "m⁻¹"],
        "followed_by": None,
    },
    "km": {
        "dim": "length",
        "degree": 1,
        "scale": 1000.0,
        "terms": [
            "kmetre",
            "kilometre",
            "kilo-metre",
            "kmetres",
            "kilometres",
            "kilo-metres",
            "km",
            "kilom",
            "kilo-m",
        ],
        "followed_by": None,
    },
    "hm": {
        "dim": "length",
        "degree": 1,
        "scale": 100.0,
        "terms": [
            "hmetre",
            "hectometre",
            "hecto-metre",
            "hmetres",
            "hectometres",
            "hecto-metres",
            "hm",
            "hectom",
            "hecto-m",
        ],
        "followed_by": None,
    },
    "dam": {
        "dim": "length",
        "degree": 1,
        "scale": 10.0,
        "terms": [
            "dametre",
            "decametre",
            "deca-metre",
            "dametres",
            "decametres",
            "deca-metres",
            "dam",
            "decam",
            "deca-m",
        ],
        "followed_by": None,
    },
    "dm": {
        "dim": "length",
        "degree": 1,
        "scale": 0.1,
        "terms": [
            "dmetre",
            "decimetre",
            "deci-metre",
            "dmetres",
            "decimetres",
            "deci-metres",
            "dm",
            "decim",
            "deci-m",
        ],
        "followed_by": None,
    },
    "cm": {
        "dim": "length",
        "degree": 1,
        "scale": 0.01,
        "terms": [
            "cmetre",
            "centimetre",
            "centi-metre",
            "cmetres",
            "centimetres",
            "centi-metres",
            "cm",
            "centim",
            "centi-m",
        ],
        "followed_by": None,
    },
    "mm": {
        "dim": "length",
        "degree": 1,
        "scale": 0.001,
        "terms": [
            "mmetre",
            "millimetre",
            "milli-metre",
            "milimetre",
            "mili-metre",
            "mmetres",
            "millimetres",
            "milli-metres",
            "milimetres",
            "mili-metres",
            "mm",
            "millim",
            "milli-m",
            "milim",
            "mili-m",
        ],
        "followed_by": None,
    },
    "μm": {
        "dim": "length",
        "degree": 1,
        "scale": 1e-06,
        "terms": [
            "μmetre",
            "umetre",
            "µmetre",
            "micrometre",
            "micro-metre",
            "μmetres",
            "umetres",
            "µmetres",
            "micrometres",
            "micro-metres",
            "μm",
            "um",
            "µm",
            "microm",
            "micro-m",
        ],
        "followed_by": None,
    },
    "nm": {
        "dim": "length",
        "degree": 1,
        "scale": 1e-09,
        "terms": [
            "nmetre",
            "nanometre",
            "nano-metre",
            "nmetres",
            "nanometres",
            "nano-metres",
            "nm",
            "nanom",
            "nano-m",
        ],
        "followed_by": None,
    },
    "pm": {
        "dim": "length",
        "degree": 1,
        "scale": 1e-12,
        "terms": [
            "pmetre",
            "picometre",
            "pico-metre",
            "pmetres",
            "picometres",
            "pico-metres",
            "pm",
            "picom",
            "pico-m",
        ],
        "followed_by": None,
    },
    "fm": {
        "dim": "length",
        "degree": 1,
        "scale": 1e-15,
        "terms": [
            "fmetre",
            "femtometre",
            "femto-metre",
            "fmetres",
            "femtometres",
            "femto-metres",
            "fm",
            "femtom",
            "femto-m",
        ],
        "followed_by": None,
    },
    "m": {
        "dim": "length",
        "degree": 1,
        "scale": 1,
        "terms": ["metre", "metres", "m"],
        "followed_by": "cm",
    },
    "per_km2": {
        "dim": "length",
        "degree": -2,
        "scale": 0.001,
        "terms": ["km-2", "kilom-2", "kilo-m-2", "km⁻²", "kilom⁻²", "kilo-m⁻²"],
        "followed_by": None,
    },
    "per_hm2": {
        "dim": "length",
        "degree": -2,
        "scale": 0.01,
        "terms": ["hm-2", "hectom-2", "hecto-m-2", "hm⁻²", "hectom⁻²", "hecto-m⁻²"],
        "followed_by": None,
    },
    "per_dam2": {
        "dim": "length",
        "degree": -2,
        "scale": 0.1,
        "terms": ["dam-2", "decam-2", "deca-m-2", "dam⁻²", "decam⁻²", "deca-m⁻²"],
        "followed_by": None,
    },
    "per_dm2": {
        "dim": "length",
        "degree": -2,
        "scale": 10.0,
        "terms": ["dm-2", "decim-2", "deci-m-2", "dm⁻²", "decim⁻²", "deci-m⁻²"],
        "followed_by": None,
    },
    "per_cm2": {
        "dim": "length",
        "degree": -2,
        "scale": 100.0,
        "terms": ["cm-2", "centim-2", "centi-m-2", "cm⁻²", "centim⁻²", "centi-m⁻²"],
        "followed_by": None,
    },
    "per_mm2": {
        "dim": "length",
        "degree": -2,
        "scale": 1000.0,
        "terms": [
            "mm-2",
            "millim-2",
            "milli-m-2",
            "milim-2",
            "mili-m-2",
            "mm⁻²",
            "millim⁻²",
            "milli-m⁻²",
            "milim⁻²",
            "mili-m⁻²",
        ],
        "followed_by": None,
    },
    "per_μm2": {
        "dim": "length",
        "degree": -2,
        "scale": 1000000.0,
        "terms": [
            "μm-2",
            "um-2",
            "µm-2",
            "microm-2",
            "micro-m-2",
            "μm⁻²",
            "um⁻²",
            "µm⁻²",
            "microm⁻²",
            "micro-m⁻²",
        ],
        "followed_by": None,
    },
    "per_nm2": {
        "dim": "length",
        "degree": -2,
        "scale": 1e9,
        "terms": ["nm-2", "nanom-2", "nano-m-2", "nm⁻²", "nanom⁻²", "nano-m⁻²"],
        "followed_by": None,
    },
    "per_pm2": {
        "dim": "length",
        "degree": -2,
        "scale": 1e12,
        "terms": ["pm-2", "picom-2", "pico-m-2", "pm⁻²", "picom⁻²", "pico-m⁻²"],
        "followed_by": None,
    },
    "per_fm2": {
        "dim": "length",
        "degree": -2,
        "scale": 1e15,
        "terms": ["fm-2", "femtom-2", "femto-m-2", "fm⁻²", "femtom⁻²", "femto-m⁻²"],
        "followed_by": None,
    },
    "per_m2": {
        "dim": "length",
        "degree": -2,
        "scale": 1.0,
        "terms": ["m-2", "m⁻²"],
        "followed_by": None,
    },
    "km2": {
        "dim": "length",
        "degree": 2,
        "scale": 1000.0,
        "terms": ["km2", "kilom2", "kilo-m2", "km²", "kilom²", "kilo-m²"],
        "followed_by": None,
    },
    "hm2": {
        "dim": "length",
        "degree": 2,
        "scale": 100.0,
        "terms": ["hm2", "hectom2", "hecto-m2", "hm²", "hectom²", "hecto-m²"],
        "followed_by": None,
    },
    "dam2": {
        "dim": "length",
        "degree": 2,
        "scale": 10.0,
        "terms": ["dam2", "decam2", "deca-m2", "dam²", "decam²", "deca-m²"],
        "followed_by": None,
    },
    "dm2": {
        "dim": "length",
        "degree": 2,
        "scale": 0.1,
        "terms": ["dm2", "decim2", "deci-m2", "dm²", "decim²", "deci-m²"],
        "followed_by": None,
    },
    "cm2": {
        "dim": "length",
        "degree": 2,
        "scale": 0.01,
        "terms": ["cm2", "centim2", "centi-m2", "cm²", "centim²", "centi-m²"],
        "followed_by": None,
    },
    "mm2": {
        "dim": "length",
        "degree": 2,
        "scale": 0.001,
        "terms": [
            "mm2",
            "millim2",
            "milli-m2",
            "milim2",
            "mili-m2",
            "mm²",
            "millim²",
            "milli-m²",
            "milim²",
            "mili-m²",
        ],
        "followed_by": None,
    },
    "μm2": {
        "dim": "length",
        "degree": 2,
        "scale": 1e-06,
        "terms": [
            "μm2",
            "um2",
            "µm2",
            "microm2",
            "micro-m2",
            "μm²",
            "um²",
            "µm²",
            "microm²",
            "micro-m²",
        ],
        "followed_by": None,
    },
    "nm2": {
        "dim": "length",
        "degree": 2,
        "scale": 1e-09,
        "terms": ["nm2", "nanom2", "nano-m2", "nm²", "nanom²", "nano-m²"],
        "followed_by": None,
    },
    "pm2": {
        "dim": "length",
        "degree": 2,
        "scale": 1e-12,
        "terms": ["pm2", "picom2", "pico-m2", "pm²", "picom²", "pico-m²"],
        "followed_by": None,
    },
    "fm2": {
        "dim": "length",
        "degree": 2,
        "scale": 1e-15,
        "terms": ["fm2", "femtom2", "femto-m2", "fm²", "femtom²", "femto-m²"],
        "followed_by": None,
    },
    "m2": {
        "dim": "length",
        "degree": 2,
        "scale": 1,
        "terms": ["m2", "m²"],
        "followed_by": None,
    },
    "per_km3": {
        "dim": "length",
        "degree": -3,
        "scale": 0.001,
        "terms": ["km-3", "kilom-3", "kilo-m-3", "km⁻³", "kilom⁻³", "kilo-m⁻³"],
        "followed_by": None,
    },
    "per_hm3": {
        "dim": "length",
        "degree": -3,
        "scale": 0.01,
        "terms": ["hm-3", "hectom-3", "hecto-m-3", "hm⁻³", "hectom⁻³", "hecto-m⁻³"],
        "followed_by": None,
    },
    "per_dam3": {
        "dim": "length",
        "degree": -3,
        "scale": 0.1,
        "terms": ["dam-3", "decam-3", "deca-m-3", "dam⁻³", "decam⁻³", "deca-m⁻³"],
        "followed_by": None,
    },
    "per_dm3": {
        "dim": "length",
        "degree": -3,
        "scale": 10.0,
        "terms": ["dm-3", "decim-3", "deci-m-3", "dm⁻³", "decim⁻³", "deci-m⁻³"],
        "followed_by": None,
    },
    "per_cm3": {
        "dim": "length",
        "degree": -3,
        "scale": 100.0,
        "terms": ["cm-3", "centim-3", "centi-m-3", "cm⁻³", "centim⁻³", "centi-m⁻³"],
        "followed_by": None,
    },
    "per_mm3": {
        "dim": "length",
        "degree": -3,
        "scale": 1000.0,
        "terms": [
            "mm-3",
            "millim-3",
            "milli-m-3",
            "milim-3",
            "mili-m-3",
            "mm⁻³",
            "millim⁻³",
            "milli-m⁻³",
            "milim⁻³",
            "mili-m⁻³",
        ],
        "followed_by": None,
    },
    "per_μm3": {
        "dim": "length",
        "degree": -3,
        "scale": 1000000.0,
        "terms": [
            "μm-3",
            "um-3",
            "µm-3",
            "microm-3",
            "micro-m-3",
            "μm⁻³",
            "um⁻³",
            "µm⁻³",
            "microm⁻³",
            "micro-m⁻³",
        ],
        "followed_by": None,
    },
    "per_nm3": {
        "dim": "length",
        "degree": -3,
        "scale": 1e9,
        "terms": ["nm-3", "nanom-3", "nano-m-3", "nm⁻³", "nanom⁻³", "nano-m⁻³"],
        "followed_by": None,
    },
    "per_pm3": {
        "dim": "length",
        "degree": -3,
        "scale": 1e12,
        "terms": ["pm-3", "picom-3", "pico-m-3", "pm⁻³", "picom⁻³", "pico-m⁻³"],
        "followed_by": None,
    },
    "per_fm3": {
        "dim": "length",
        "degree": -3,
        "scale": 1e15,
        "terms": ["fm-3", "femtom-3", "femto-m-3", "fm⁻³", "femtom⁻³", "femto-m⁻³"],
        "followed_by": None,
    },
    "per_m3": {
        "dim": "length",
        "degree": -3,
        "scale": 1.0,
        "terms": ["m-3", "m⁻³"],
        "followed_by": None,
    },
    "km3": {
        "dim": "length",
        "degree": 3,
        "scale": 1000.0,
        "terms": ["km3", "kilom3", "kilo-m3", "km³", "kilom³", "kilo-m³"],
        "followed_by": None,
    },
    "hm3": {
        "dim": "length",
        "degree": 3,
        "scale": 100.0,
        "terms": ["hm3", "hectom3", "hecto-m3", "hm³", "hectom³", "hecto-m³"],
        "followed_by": None,
    },
    "dam3": {
        "dim": "length",
        "degree": 3,
        "scale": 10.0,
        "terms": ["dam3", "decam3", "deca-m3", "dam³", "decam³", "deca-m³"],
        "followed_by": None,
    },
    "dm3": {
        "dim": "length",
        "degree": 3,
        "scale": 0.1,
        "terms": ["dm3", "decim3", "deci-m3", "dm³", "decim³", "deci-m³"],
        "followed_by": None,
    },
    "cm3": {
        "dim": "length",
        "degree": 3,
        "scale": 0.01,
        "terms": ["cm3", "centim3", "centi-m3", "cm³", "centim³", "centi-m³"],
        "followed_by": None,
    },
    "mm3": {
        "dim": "length",
        "degree": 3,
        "scale": 0.001,
        "terms": [
            "mm3",
            "millim3",
            "milli-m3",
            "milim3",
            "mili-m3",
            "mm³",
            "millim³",
            "milli-m³",
            "milim³",
            "mili-m³",
        ],
        "followed_by": None,
    },
    "μm3": {
        "dim": "length",
        "degree": 3,
        "scale": 1e-06,
        "terms": [
            "μm3",
            "um3",
            "µm3",
            "microm3",
            "micro-m3",
            "μm³",
            "um³",
            "µm³",
            "microm³",
            "micro-m³",
        ],
        "followed_by": None,
    },
    "nm3": {
        "dim": "length",
        "degree": 3,
        "scale": 1e-09,
        "terms": ["nm3", "nanom3", "nano-m3", "nm³", "nanom³", "nano-m³"],
        "followed_by": None,
    },
    "pm3": {
        "dim": "length",
        "degree": 3,
        "scale": 1e-12,
        "terms": ["pm3", "picom3", "pico-m3", "pm³", "picom³", "pico-m³"],
        "followed_by": None,
    },
    "fm3": {
        "dim": "length",
        "degree": 3,
        "scale": 1e-15,
        "terms": ["fm3", "femtom3", "femto-m3", "fm³", "femtom³", "femto-m³"],
        "followed_by": None,
    },
    "m3": {
        "dim": "length",
        "degree": 3,
        "scale": 1,
        "terms": ["m3", "m³"],
        "followed_by": None,
    },
    "kmol": {
        "dim": "mole",
        "degree": 1,
        "scale": 1000.0,
        "terms": [
            "kmol",
            "kilomol",
            "kilo-mol",
            "kmole",
            "kilomole",
            "kilo-mole",
            "kmoles",
            "kilomoles",
            "kilo-moles",
        ],
        "followed_by": None,
    },
    "hmol": {
        "dim": "mole",
        "degree": 1,
        "scale": 100.0,
        "terms": [
            "hmol",
            "hectomol",
            "hecto-mol",
            "hmole",
            "hectomole",
            "hecto-mole",
            "hmoles",
            "hectomoles",
            "hecto-moles",
        ],
        "followed_by": None,
    },
    "damol": {
        "dim": "mole",
        "degree": 1,
        "scale": 10.0,
        "terms": [
            "damol",
            "decamol",
            "deca-mol",
            "damole",
            "decamole",
            "deca-mole",
            "damoles",
            "decamoles",
            "deca-moles",
        ],
        "followed_by": None,
    },
    "dmol": {
        "dim": "mole",
        "degree": 1,
        "scale": 0.1,
        "terms": [
            "dmol",
            "decimol",
            "deci-mol",
            "dmole",
            "decimole",
            "deci-mole",
            "dmoles",
            "decimoles",
            "deci-moles",
        ],
        "followed_by": None,
    },
    "cmol": {
        "dim": "mole",
        "degree": 1,
        "scale": 0.01,
        "terms": [
            "cmol",
            "centimol",
            "centi-mol",
            "cmole",
            "centimole",
            "centi-mole",
            "cmoles",
            "centimoles",
            "centi-moles",
        ],
        "followed_by": None,
    },
    "mmol": {
        "dim": "mole",
        "degree": 1,
        "scale": 0.001,
        "terms": [
            "mmol",
            "millimol",
            "milli-mol",
            "milimol",
            "mili-mol",
            "mmole",
            "millimole",
            "milli-mole",
            "milimole",
            "mili-mole",
            "mmoles",
            "millimoles",
            "milli-moles",
            "milimoles",
            "mili-moles",
        ],
        "followed_by": None,
    },
    "μmol": {
        "dim": "mole",
        "degree": 1,
        "scale": 1e-06,
        "terms": [
            "μmol",
            "umol",
            "µmol",
            "micromol",
            "micro-mol",
            "μmole",
            "umole",
            "µmole",
            "micromole",
            "micro-mole",
            "μmoles",
            "umoles",
            "µmoles",
            "micromoles",
            "micro-moles",
        ],
        "followed_by": None,
    },
    "nmol": {
        "dim": "mole",
        "degree": 1,
        "scale": 1e-09,
        "terms": [
            "nmol",
            "nanomol",
            "nano-mol",
            "nmole",
            "nanomole",
            "nano-mole",
            "nmoles",
            "nanomoles",
            "nano-moles",
        ],
        "followed_by": None,
    },
    "pmol": {
        "dim": "mole",
        "degree": 1,
        "scale": 1e-12,
        "terms": [
            "pmol",
            "picomol",
            "pico-mol",
            "pmole",
            "picomole",
            "pico-mole",
            "pmoles",
            "picomoles",
            "pico-moles",
        ],
        "followed_by": None,
    },
    "fmol": {
        "dim": "mole",
        "degree": 1,
        "scale": 1e-15,
        "terms": [
            "fmol",
            "femtomol",
            "femto-mol",
            "fmole",
            "femtomole",
            "femto-mole",
            "fmoles",
            "femtomoles",
            "femto-moles",
        ],
        "followed_by": None,
    },
    "mol": {
        "dim": "mole",
        "degree": 1,
        "scale": 1,
        "terms": ["mol", "mole", "moles"],
        "followed_by": None,
    },
    "per_kui": {
        "dim": "ui",
        "degree": -1,
        "scale": 0.001,
        "terms": ["kui-1", "kiloui-1", "kilo-ui-1", "kui⁻¹", "kiloui⁻¹", "kilo-ui⁻¹"],
        "followed_by": None,
    },
    "per_hui": {
        "dim": "ui",
        "degree": -1,
        "scale": 0.01,
        "terms": [
            "hui-1",
            "hectoui-1",
            "hecto-ui-1",
            "hui⁻¹",
            "hectoui⁻¹",
            "hecto-ui⁻¹",
        ],
        "followed_by": None,
    },
    "per_daui": {
        "dim": "ui",
        "degree": -1,
        "scale": 0.1,
        "terms": ["daui-1", "decaui-1", "deca-ui-1", "daui⁻¹", "decaui⁻¹", "deca-ui⁻¹"],
        "followed_by": None,
    },
    "per_dui": {
        "dim": "ui",
        "degree": -1,
        "scale": 10.0,
        "terms": ["dui-1", "deciui-1", "deci-ui-1", "dui⁻¹", "deciui⁻¹", "deci-ui⁻¹"],
        "followed_by": None,
    },
    "per_cui": {
        "dim": "ui",
        "degree": -1,
        "scale": 100.0,
        "terms": [
            "cui-1",
            "centiui-1",
            "centi-ui-1",
            "cui⁻¹",
            "centiui⁻¹",
            "centi-ui⁻¹",
        ],
        "followed_by": None,
    },
    "per_mui": {
        "dim": "ui",
        "degree": -1,
        "scale": 1000.0,
        "terms": [
            "mui-1",
            "milliui-1",
            "milli-ui-1",
            "miliui-1",
            "mili-ui-1",
            "mui⁻¹",
            "milliui⁻¹",
            "milli-ui⁻¹",
            "miliui⁻¹",
            "mili-ui⁻¹",
        ],
        "followed_by": None,
    },
    "per_μui": {
        "dim": "ui",
        "degree": -1,
        "scale": 1000000.0,
        "terms": [
            "μui-1",
            "uui-1",
            "µui-1",
            "microui-1",
            "micro-ui-1",
            "μui⁻¹",
            "uui⁻¹",
            "µui⁻¹",
            "microui⁻¹",
            "micro-ui⁻¹",
        ],
        "followed_by": None,
    },
    "per_nui": {
        "dim": "ui",
        "degree": -1,
        "scale": 1e9,
        "terms": ["nui-1", "nanoui-1", "nano-ui-1", "nui⁻¹", "nanoui⁻¹", "nano-ui⁻¹"],
        "followed_by": None,
    },
    "per_pui": {
        "dim": "ui",
        "degree": -1,
        "scale": 1e12,
        "terms": ["pui-1", "picoui-1", "pico-ui-1", "pui⁻¹", "picoui⁻¹", "pico-ui⁻¹"],
        "followed_by": None,
    },
    "per_fui": {
        "dim": "ui",
        "degree": -1,
        "scale": 1e15,
        "terms": [
            "fui-1",
            "femtoui-1",
            "femto-ui-1",
            "fui⁻¹",
            "femtoui⁻¹",
            "femto-ui⁻¹",
        ],
        "followed_by": None,
    },
    "per_ui": {
        "dim": "ui",
        "degree": -1,
        "scale": 1.0,
        "terms": ["ui-1", "ui⁻¹"],
        "followed_by": None,
    },
    "kui": {
        "dim": "ui",
        "degree": 1,
        "scale": 1000.0,
        "terms": ["kui", "kiloui", "kilo-ui", "ku", "kilou", "kilo-u"],
        "followed_by": None,
    },
    "hui": {
        "dim": "ui",
        "degree": 1,
        "scale": 100.0,
        "terms": ["hui", "hectoui", "hecto-ui", "hu", "hectou", "hecto-u"],
        "followed_by": None,
    },
    "daui": {
        "dim": "ui",
        "degree": 1,
        "scale": 10.0,
        "terms": ["daui", "decaui", "deca-ui", "dau", "decau", "deca-u"],
        "followed_by": None,
    },
    "dui": {
        "dim": "ui",
        "degree": 1,
        "scale": 0.1,
        "terms": ["dui", "deciui", "deci-ui", "du", "deciu", "deci-u"],
        "followed_by": None,
    },
    "cui": {
        "dim": "ui",
        "degree": 1,
        "scale": 0.01,
        "terms": ["cui", "centiui", "centi-ui", "cu", "centiu", "centi-u"],
        "followed_by": None,
    },
    "mui": {
        "dim": "ui",
        "degree": 1,
        "scale": 0.001,
        "terms": [
            "mui",
            "milliui",
            "milli-ui",
            "miliui",
            "mili-ui",
            "mu",
            "milliu",
            "milli-u",
            "miliu",
            "mili-u",
        ],
        "followed_by": None,
    },
    "μui": {
        "dim": "ui",
        "degree": 1,
        "scale": 1e-06,
        "terms": [
            "μui",
            "uui",
            "µui",
            "microui",
            "micro-ui",
            "μu",
            "uu",
            "µu",
            "microu",
            "micro-u",
        ],
        "followed_by": None,
    },
    "nui": {
        "dim": "ui",
        "degree": 1,
        "scale": 1e-09,
        "terms": ["nui", "nanoui", "nano-ui", "nu", "nanou", "nano-u"],
        "followed_by": None,
    },
    "pui": {
        "dim": "ui",
        "degree": 1,
        "scale": 1e-12,
        "terms": ["pui", "picoui", "pico-ui", "pu", "picou", "pico-u"],
        "followed_by": None,
    },
    "fui": {
        "dim": "ui",
        "degree": 1,
        "scale": 1e-15,
        "terms": ["fui", "femtoui", "femto-ui", "fu", "femtou", "femto-u"],
        "followed_by": None,
    },
    "ui": {
        "dim": "ui",
        "degree": 1,
        "scale": 1,
        "terms": ["ui", "u"],
        "followed_by": None,
    },
    "kPa": {
        "dim": "pressure",
        "degree": 1,
        "scale": 1000.0,
        "terms": ["kPa", "kiloPa", "kilo-Pa"],
        "followed_by": None,
    },
    "hPa": {
        "dim": "pressure",
        "degree": 1,
        "scale": 100.0,
        "terms": ["hPa", "hectoPa", "hecto-Pa"],
        "followed_by": None,
    },
    "daPa": {
        "dim": "pressure",
        "degree": 1,
        "scale": 10.0,
        "terms": ["daPa", "decaPa", "deca-Pa"],
        "followed_by": None,
    },
    "dPa": {
        "dim": "pressure",
        "degree": 1,
        "scale": 0.1,
        "terms": ["dPa", "deciPa", "deci-Pa"],
        "followed_by": None,
    },
    "cPa": {
        "dim": "pressure",
        "degree": 1,
        "scale": 0.01,
        "terms": ["cPa", "centiPa", "centi-Pa"],
        "followed_by": None,
    },
    "mPa": {
        "dim": "pressure",
        "degree": 1,
        "scale": 0.001,
        "terms": ["mPa", "milliPa", "milli-Pa", "miliPa", "mili-Pa"],
        "followed_by": None,
    },
    "μPa": {
        "dim": "pressure",
        "degree": 1,
        "scale": 1e-06,
        "terms": ["μPa", "uPa", "µPa", "microPa", "micro-Pa"],
        "followed_by": None,
    },
    "nPa": {
        "dim": "pressure",
        "degree": 1,
        "scale": 1e-09,
        "terms": ["nPa", "nanoPa", "nano-Pa"],
        "followed_by": None,
    },
    "pPa": {
        "dim": "pressure",
        "degree": 1,
        "scale": 1e-12,
        "terms": ["pPa", "picoPa", "pico-Pa"],
        "followed_by": None,
    },
    "fPa": {
        "dim": "pressure",
        "degree": 1,
        "scale": 1e-15,
        "terms": ["fPa", "femtoPa", "femto-Pa"],
        "followed_by": None,
    },
    "Pa": {
        "dim": "pressure",
        "degree": 1,
        "scale": 1,
        "terms": ["Pa"],
        "followed_by": None,
    },
    "%": {"dim": "%", "degree": 1, "scale": 1, "terms": ["%"], "followed_by": None},
    "log": {
        "dim": "log",
        "degree": 1,
        "scale": 1,
        "terms": ["log"],
        "followed_by": None,
    },
    "mmHg": {
        "dim": "pressure",
        "degree": 1,
        "scale": 133,
        "terms": ["mmHg"],
        "followed_by": None,
    },
    "second": {
        "dim": "time",
        "degree": 1,
        "scale": 1,
        "terms": ["seconde", "secondes", "s"],
        "followed_by": None,
    },
    "minute": {
        "dim": "time",
        "degree": 1,
        "scale": 60,
        "terms": ["mn", "min", "minute", "minutes"],
        "followed_by": "second",
    },
    "hour": {
        "dim": "time",
        "degree": 1,
        "scale": 3600,
        "terms": ["heure", "heures", "h"],
        "followed_by": "minute",
    },
    "day": {
        "dim": "time",
        "degree": 1,
        "scale": 86400,
        "terms": ["jour", "jours", "j"],
        "followed_by": None,
    },
    "month": {
        "dim": "time",
        "degree": 1,
        "scale": 2628002.88,
        "terms": ["mois"],
        "followed_by": None,
    },
    "week": {
        "dim": "time",
        "degree": 1,
        "scale": 604800,
        "terms": ["semaine", "semaines", "sem"],
        "followed_by": None,
    },
    "year": {
        "dim": "time",
        "degree": 1,
        "scale": 31557600.0,
        "terms": ["an", "année", "ans", "années"],
        "followed_by": None,
    },
    "arc-second": {
        "dim": "time",
        "degree": 1,
        "scale": 0.03333333333333333,
        "terms": ['"', "''"],
        "followed_by": None,
    },
    "arc-minute": {
        "dim": "time",
        "degree": 1,
        "scale": 2,
        "terms": ["'"],
        "followed_by": "arc-second",
    },
    "degree": {
        "dim": "time",
        "degree": 1,
        "scale": 120,
        "terms": ["degre", "°", "deg"],
        "followed_by": "arc-minute",
    },
    "celsius": {
        "dim": "temperature",
        "degree": 1,
        "scale": 1,
        "terms": ["°C", "° celsius", "celsius"],
        "followed_by": None,
    },
    "cac": {
        "dim": "length",
        "degree": 3,
        "scale": 5.0,
        "terms": ["cac", "c.a.c", "cuillere à café", "cuillères à café"],
        "followed_by": None,
    },
    "goutte": {
        "dim": "length",
        "degree": 3,
        "scale": 0.05,
        "terms": ["gt", "goutte", "gouttes"],
        "followed_by": None,
    },
    "ml": {
        "dim": "length",
        "degree": 3,
        "scale": 1.0,
        "terms": ["mililitre", "millilitre", "mililitres", "millilitres", "ml"],
        "followed_by": None,
    },
    "cl": {
        "dim": "length",
        "degree": 3,
        "scale": 10.0,
        "terms": ["centilitre", "centilitres", "cl"],
        "followed_by": None,
    },
    "dl": {
        "dim": "length",
        "degree": 3,
        "scale": 100.0,
        "terms": ["decilitre", "decilitres", "dl"],
        "followed_by": None,
    },
    "l": {
        "dim": "length",
        "degree": 3,
        "scale": 1000.0,
        "terms": ["litre", "litres", "l", "dm3"],
        "followed_by": "ml",
    },
    "x10*1": {
        "dim": "count",
        "degree": 0,
        "scale": 10,
        "terms": ["x10*1"],
        "followed_by": None,
    },
    "x10*2": {
        "dim": "count",
        "degree": 0,
        "scale": 100,
        "terms": ["x10*2"],
        "followed_by": None,
    },
    "x10*3": {
        "dim": "count",
        "degree": 0,
        "scale": 1000,
        "terms": ["x10*3"],
        "followed_by": None,
    },
    "x10*4": {
        "dim": "count",
        "degree": 0,
        "scale": 10000,
        "terms": ["x10*4"],
        "followed_by": None,
    },
    "x10*5": {
        "dim": "count",
        "degree": 0,
        "scale": 100000,
        "terms": ["x10*5"],
        "followed_by": None,
    },
    "x10*6": {
        "dim": "count",
        "degree": 0,
        "scale": 1000000,
        "terms": ["x10*6"],
        "followed_by": None,
    },
    "x10*7": {
        "dim": "count",
        "degree": 0,
        "scale": 10000000,
        "terms": ["x10*7"],
        "followed_by": None,
    },
    "x10*8": {
        "dim": "count",
        "degree": 0,
        "scale": 100000000,
        "terms": ["x10*8"],
        "followed_by": None,
    },
    "x10*9": {
        "dim": "count",
        "degree": 0,
        "scale": 1000000000,
        "terms": ["x10*9"],
        "followed_by": None,
    },
    "x10*10": {
        "dim": "count",
        "degree": 0,
        "scale": 10000000000,
        "terms": ["x10*10"],
        "followed_by": None,
    },
    "x10*11": {
        "dim": "count",
        "degree": 0,
        "scale": 100000000000,
        "terms": ["x10*11"],
        "followed_by": None,
    },
    "x10*12": {
        "dim": "count",
        "degree": 0,
        "scale": 1000000000000,
        "terms": ["x10*12"],
        "followed_by": None,
    },
    "x10*13": {
        "dim": "count",
        "degree": 0,
        "scale": 10000000000000,
        "terms": ["x10*13"],
        "followed_by": None,
    },
    "x10*14": {
        "dim": "count",
        "degree": 0,
        "scale": 100000000000000,
        "terms": ["x10*14"],
        "followed_by": None,
    },
    "x10*15": {
        "dim": "count",
        "degree": 0,
        "scale": 1000000000000000,
        "terms": ["x10*15"],
        "followed_by": None,
    },
}



import itertools
import json
import math
import random
import time
from collections import defaultdict
from collections.abc import Sized
from itertools import chain, repeat
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Optional,
    Tuple,
)

import spacy
import spacy.tokenizer
import torch
from accelerate import Accelerator
from confit import Cli
from confit.registry import validate_arguments
from confit.utils.random import set_seed
from pydantic import BaseModel
from rich_logger import RichTablePrinter
from spacy.tokens import Doc
from tqdm import tqdm

import edsnlp
from edsnlp.core.pipeline import Pipeline
from edsnlp.core.registries import registry
from edsnlp.optimization import LinearSchedule, ScheduledOptimizer
from edsnlp.pipes.base import BaseNERComponent, BaseSpanAttributeClassifierComponent
from edsnlp.pipes.trainable.embeddings.transformer.transformer import Transformer
from edsnlp.scorers.ner import NerScorer
from edsnlp.scorers.span_attributes import SpanAttributeScorer
from edsnlp.utils.bindings import BINDING_SETTERS
from edsnlp.utils.collections import batchify
from edsnlp.utils.span_getters import get_spans
from edsnlp.utils.typing import AsList

app = Cli(pretty_exceptions_show_locals=False)

LOGGER_FIELDS = {
    "step": {},
    "(.*)loss": {
        "goal": "lower_is_better",
        "format": "{:.2e}",
        "goal_wait": 2,
    },
    "lr": {"format": "{:.2e}"},
    "speed/(.*)": {"format": "{:.2f}", r"name": r"\1"},
    "labels": {"format": "{:.2f}"},
    "(.*?)/micro/(f|r|p)$": {
        "goal": "higher_is_better",
        "format": "{:.2%}",
        "goal_wait": 1,
        "name": r"\1_\2",
    },
}


def flatten_dict(d, path=""):
    if not isinstance(d, dict):
        return {path: d}

    return {
        k: v
        for key, val in d.items()
        for k, v in flatten_dict(val, f"{path}/{key}" if path else key).items()
    }


class BatchSizeArg:
    """
    Batch size argument validator / caster for confit/pydantic

    Examples
    --------

    ```{ .python .no-check }
    def fn(batch_size: BatchSizeArg):
        return batch_size


    print(fn("10 samples"))
    # Out: (10, "samples")

    print(fn("10 words"))
    # Out: (10, "words")

    print(fn(10))
    # Out: (10, "samples")
    ```
    """

    @classmethod
    def validate(cls, value, config=None):
        value = str(value)
        parts = value.split()
        num = int(parts[0])
        unit = parts[1] if len(parts) == 2 else "samples"
        if (
            len(parts) == 2
            and str(num) == parts[0]
            and unit in ("words", "samples", "spans")
        ):
            return num, unit
        raise Exception(
            f"Invalid batch size: {value}, must be <int> samples|words|spans"
        )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate


if TYPE_CHECKING:
    BatchSizeArg = Tuple[int, str]  # noqa: F811


class LengthSortedBatchSampler:
    """
    Batch sampler that sorts the dataset by length and then batches
    sequences of similar length together. This is useful for transformer
    models that can then be padded more efficiently.

    Parameters
    ----------
    dataset: Iterable
        The dataset to sample from (can be a generator or a fixed size collection)
    batch_size: int
        The batch size
    batch_unit: str
        The unit of the batch size, either "words" or "samples"
    noise: int
        The amount of noise to add to the sequence length before sorting
        (uniformly sampled in [-noise, noise])
    drop_last: bool
        Whether to drop the last batch if it is smaller than the batch size
    buffer_size: Optional[int]
        The size of the buffer to use to shuffle the batches. If None, the buffer
        will be approximately the size of the dataset.
    """

    def __init__(
        self,
        dataset,
        batch_size: int,
        batch_unit: str,
        noise=1,
        drop_last=True,
        buffer_size: Optional[int] = None,
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.batch_unit = batch_unit
        self.noise = noise
        self.drop_last = drop_last
        self.buffer_size = buffer_size

    def __iter__(self):
        # Shuffle the dataset
        if self.batch_unit == "words":

            def sample_len(idx, noise=True):
                count = sum(
                    len(x)
                    for x in next(
                        v
                        for k, v in self.dataset[idx].items()
                        if k.endswith("word_lengths")
                    )
                )
                if not noise:
                    return count
                return count + random.randint(-self.noise, self.noise)

        elif self.batch_unit == "spans":

            def sample_len(idx, noise=True):
                return len(
                    next(
                        v for k, v in self.dataset[idx].items() if k.endswith("begins")
                    )
                )

        else:
            sample_len = lambda idx, noise=True: 1  # noqa: E731

        def make_batches():
            total = 0
            batch = []
            for seq_size, idx in sorted_sequences:
                if total and total + seq_size > self.batch_size:
                    yield batch
                    total = 0
                    batch = []
                total += seq_size
                batch.append(idx)

        # Shuffle the batches in buffer that contain approximately
        # the full dataset to add more randomness
        if isinstance(self.dataset, Sized):
            total_count = sum(sample_len(i, False) for i in range(len(self.dataset)))

        assert (
            isinstance(self.dataset, Sized) or self.buffer_size is not None
        ), "Dataset must have a length or buffer_size must be specified"
        buffer_size = self.buffer_size or math.ceil(total_count / self.batch_size)

        # Sort sequences by length +- some noise
        sorted_sequences = chain.from_iterable(
            sorted((sample_len(i), i) for i in range(len(self.dataset)))
            for _ in repeat(None)
        )

        # Batch sorted sequences
        batches = make_batches()
        buffers = batchify(batches, buffer_size)
        for buffer in buffers:
            random.shuffle(buffer)
            yield from buffer


class SubBatchCollater:
    """
    Collater that splits batches into sub-batches of a maximum size

    Parameters
    ----------
    nlp: Pipeline
        The pipeline object
    embedding: Transformer
        The transformer embedding pipe
    grad_accumulation_max_tokens: int
        The maximum number of tokens (word pieces) to accumulate in a single batch
    """

    def __init__(self, nlp, embedding, grad_accumulation_max_tokens):
        self.nlp = nlp
        self.embedding: Transformer = embedding
        self.grad_accumulation_max_tokens = grad_accumulation_max_tokens

    def __call__(self, seq):
        total = 0
        mini_batches = [[]]
        for sample_features in seq:
            num_tokens = sum(
                math.ceil(len(p) / self.embedding.stride) * self.embedding.window
                for key in sample_features
                if key.endswith("/input_ids")
                for p in sample_features[key]
            )
            if total + num_tokens > self.grad_accumulation_max_tokens:
                print(
                    f"Mini batch size was becoming too large: {total} > "
                    f"{self.grad_accumulation_max_tokens} so it was split"
                )
                total = 0
                mini_batches.append([])
            total += num_tokens
            mini_batches[-1].append(sample_features)
        return [self.nlp.collate(b) for b in mini_batches]


def subset_doc(doc: Doc, start: int, end: int) -> Doc:
    """
    Subset a doc given a start and end index.

    Parameters
    ----------
    doc: Doc
        The doc to subset
    start: int
        The start index
    end: int
        The end index

    Returns
    -------
    Doc
    """
    # TODO: review user_data copy strategy
    new_doc = doc[start:end].as_doc(copy_user_data=True)
    new_doc.user_data.update(doc.user_data)

    for name, group in doc.spans.items():
        new_doc.spans[name] = [
            spacy.tokens.Span(
                new_doc,
                max(0, span.start - start),
                min(end, span.end) - start,
                span.label,
            )
            for span in group
            if span.end > start and span.start < end
        ]

    return new_doc


class Reader(BaseModel, arbitrary_types_allowed=True):
    """
    Reader that reads docs from a file or a generator, and adapts them to the pipeline.

    Parameters
    ----------
    reader: Callable[..., Iterable[Doc]]
        The reader object
    limit: Optional[int]
        The maximum number of docs to read
    max_length: int
        The maximum length of the resulting docs
    randomize: bool
        Whether to randomize the split
    multi_sentence: bool
        Whether to split sentences across multiple docs
    filter_expr: Optional[str]
        An expression to filter the docs to generate
    """

    reader: Any
    limit: Optional[int] = -1
    max_length: int = 0
    randomize: bool = False
    multi_sentence: bool = True
    filter_expr: Optional[str] = None

    def __call__(self, nlp) -> List[Doc]:
        filter_fn = eval(f"lambda doc:{self.filter_expr}") if self.filter_expr else None

        blank_nlp = edsnlp.Pipeline(nlp.lang, vocab=nlp.vocab, vocab_config=None)
        if self.max_length > 0:
            blank_nlp.add_pipe("eds.sentences")

        docs = blank_nlp.pipe(self.reader).set_processing(show_progress=True)

        count = 0

        # Load the jsonl data from path
        if self.randomize:
            docs: List[Doc] = list(docs)
            random.shuffle(docs)

        for doc in docs:
            if 0 <= self.limit <= count:
                break
            if not (len(doc) and (filter_fn is None or filter_fn(doc))):
                continue
            count += 1

            for sub_doc in self.split_doc(doc):
                if len(sub_doc.text.strip()):
                    yield sub_doc
            else:
                continue

    def split_doc(
        self,
        doc: Doc,
    ) -> Iterable[Doc]:
        """
        Split a doc into multiple docs of max_length tokens.

        Parameters
        ----------
        doc: Doc
            The doc to split

        Returns
        -------
        Iterable[Doc]
        """
        max_length = self.max_length
        randomize = self.randomize

        if max_length <= 0:
            yield doc
        else:
            start = 0
            end = 0
            for ent in doc.ents:
                for token in ent:
                    token.is_sent_start = False
            for sent in doc.sents if doc.has_annotation("SENT_START") else (doc[:],):
                # If the sentence adds too many tokens
                if sent.end - start > max_length:
                    # But the current buffer too large
                    while sent.end - start > max_length:
                        subset_end = start + int(
                            max_length * (random.random() ** 0.3 if randomize else 1)
                        )
                        yield subset_doc(doc, start, subset_end)
                        start = subset_end
                    yield subset_doc(doc, start, sent.end)
                    start = sent.end

                if not self.multi_sentence:
                    yield subset_doc(doc, start, sent.end)
                    start = sent.end

                # Otherwise, extend the current buffer
                end = sent.end

            yield subset_doc(doc, start, end)


@validate_arguments
class GenericScorer:
    def __init__(self, speed=True, **scorers):
        self.scorers = scorers
        self.speed = speed

    def __call__(self, nlp: Pipeline, docs):
        scores = {}
        docs = list(docs)

        # Speed
        if self.speed:
            t0 = time.time()
            list(nlp.pipe(d.copy() for d in tqdm(docs, desc="Computing model speed")))
            duration = time.time() - t0
            scores["speed"] = dict(
                wps=sum(len(d) for d in docs) / duration,
                dps=len(docs) / duration,
            )

        # NER
        ner_pipes = [
            name for name, pipe in nlp.pipeline if isinstance(pipe, BaseNERComponent)
        ]
        ner_scorers = {
            name: scorer
            for name, scorer in self.scorers.items()
            if isinstance(scorer, NerScorer)
        }
        if ner_pipes and ner_scorers:
            clean_ner_docs = [d.copy() for d in tqdm(docs, desc="Copying docs")]
            for d in clean_ner_docs:
                d.ents = []
                d.spans.clear()
            with nlp.select_pipes(enable=ner_pipes):
                ner_preds = list(nlp.pipe(tqdm(clean_ner_docs, desc="Predicting")))
            for name, scorer in ner_scorers.items():
                scores[name] = scorer(docs, ner_preds)

        # Qualification
        qlf_pipes = [
            name
            for name, pipe in nlp.pipeline
            if isinstance(pipe, BaseSpanAttributeClassifierComponent)
        ]
        span_attr_scorers = {
            name: scorer
            for name, scorer in self.scorers.items()
            if isinstance(scorer, SpanAttributeScorer)
        }
        if qlf_pipes and span_attr_scorers:
            clean_qlf_docs = [d.copy() for d in tqdm(docs, desc="Copying docs")]
            for doc in clean_qlf_docs:
                for name in qlf_pipes:
                    pipe = nlp.get_pipe(name)
                    for span in get_spans(doc, pipe.span_getter):
                        for qlf in nlp.get_pipe(name).attributes:
                            BINDING_SETTERS[(qlf, None)](span)
            with nlp.select_pipes(disable=ner_pipes):
                qlf_preds = list(nlp.pipe(tqdm(clean_qlf_docs, desc="Predicting")))
            for name, scorer in span_attr_scorers.items():
                scores[name] = scorer(docs, qlf_preds)
        return scores


@app.command(name="train", registry=registry)
def train(
    *,
    output_path: Path = Path("model"),
    nlp: Pipeline,
    train_data: AsList[Reader],
    val_data: AsList[Reader],
    seed: int = 42,
    data_seed: int = 42,
    max_steps: int = 1000,
    batch_size: BatchSizeArg = 2000,
    transformer_lr: float = 5e-5,
    task_lr: float = 3e-4,
    validation_interval: int = 10,
    max_grad_norm: float = 5.0,
    warmup_rate: float = 0.1,
    scorer: GenericScorer = GenericScorer(),
    grad_accumulation_max_tokens: int = 96 * 128,
    cpu: bool = False,
):
    trf_pipe = next(
        module
        for name, pipe in nlp.torch_components()
        for module_name, module in pipe.named_component_modules()
        if isinstance(module, Transformer)
    )

    set_seed(seed)
    # Loading and adapting the training and validation data
    with set_seed(data_seed):
        train_docs: List[spacy.tokens.Doc] = list(
            chain.from_iterable(td(nlp) for td in train_data)
        )
        val_docs: List[spacy.tokens.Doc] = list(
            chain.from_iterable(td(nlp) for td in val_data)
        )

    # Initialize pipeline with training documents
    nlp.post_init(train_docs)

    # Preprocessing training data
    preprocessed = list(
        nlp.preprocess_many(train_docs, supervision=True).set_processing(
            show_progress=True
        )
    )
    dataloader = torch.utils.data.DataLoader(
        preprocessed,
        batch_sampler=LengthSortedBatchSampler(
            preprocessed,
            batch_size=batch_size[0],
            batch_unit=batch_size[1],
        ),
        collate_fn=SubBatchCollater(
            nlp,
            trf_pipe,
            grad_accumulation_max_tokens=grad_accumulation_max_tokens,
        ),
    )
    pipe_names, trained_pipes = zip(*nlp.torch_components())
    print("Training", ", ".join(pipe_names))

    trf_params = set(trf_pipe.parameters())
    params = set(p for pipe in trained_pipes for p in pipe.parameters())
    optimizer = ScheduledOptimizer(
        torch.optim.AdamW(
            [
                {
                    "params": list(params - trf_params),
                    "lr": task_lr,
                    "schedules": LinearSchedule(
                        total_steps=max_steps,
                        warmup_rate=warmup_rate,
                        start_value=task_lr,
                    ),
                },
                {
                    "params": list(trf_params),
                    "lr": transformer_lr,
                    "schedules": LinearSchedule(
                        total_steps=max_steps,
                        warmup_rate=warmup_rate,
                        start_value=0,
                    ),
                },
            ][: 2 if transformer_lr else 1]
        )
    )
    grad_params = {p for group in optimizer.param_groups for p in group["params"]}
    print(
        "Optimizing:"
        + "".join(
            f"\n - {len(group['params'])} params "
            f"({sum(p.numel() for p in group['params'])} total)"
            for group in optimizer.param_groups
        )
    )
    print(f"Not optimizing {len(params - grad_params)} params")
    for param in params - grad_params:
        param.requires_grad_(False)

    accelerator = Accelerator(cpu=cpu)
    print("Device:", accelerator.device)
    [dataloader, optimizer, *trained_pipes] = accelerator.prepare(
        dataloader,
        optimizer,
        *trained_pipes,
    )

    cumulated_data = defaultdict(lambda: 0.0, count=0)

    iterator = itertools.chain.from_iterable(itertools.repeat(dataloader))
    all_metrics = []
    nlp.train(True)
    set_seed(seed)

    with RichTablePrinter(LOGGER_FIELDS, auto_refresh=False) as logger:
        with tqdm(
            range(max_steps + 1),
            "Training model",
            leave=True,
            mininterval=5.0,
        ) as bar:
            for step in bar:
                if (step % validation_interval) == 0:
                    scores = scorer(nlp, val_docs)
                    all_metrics.append(
                        {
                            "step": step,
                            "lr": optimizer.param_groups[0]["lr"],
                            **cumulated_data,
                            **scores,
                        }
                    )
                    cumulated_data = defaultdict(lambda: 0.0, count=0)
                    nlp.to_disk(output_path)
                    (output_path / "train_metrics.json").write_text(
                        json.dumps(all_metrics, indent=2)
                    )
                    logger.log_metrics(flatten_dict(all_metrics[-1]))
                if step == max_steps:
                    break
                mini_batches = next(iterator)
                optimizer.zero_grad()
                for mini_batch in mini_batches:
                    loss = torch.zeros((), device=accelerator.device)
                    with nlp.cache():
                        for name, pipe in zip(pipe_names, trained_pipes):
                            output = pipe(mini_batch[name])
                            if "loss" in output:
                                loss += output["loss"]
                            for key, value in output.items():
                                if key.endswith("loss"):
                                    cumulated_data[key] += float(value)
                            if torch.isnan(loss):
                                raise ValueError(f"NaN loss at component {name}")

                    accelerator.backward(loss)

                torch.nn.utils.clip_grad_norm_(grad_params, max_grad_norm)
                optimizer.step()

    return nlp


if __name__ == "__main__":
    app()



import abc
from collections import defaultdict
from typing import Any, Dict, Optional

from edsnlp import registry
from edsnlp.scorers import Examples, make_examples, prf
from edsnlp.utils.span_getters import SpanGetter, SpanGetterArg, get_spans


def ner_exact_scorer(
    examples: Examples,
    span_getter: SpanGetter,
    micro_key: str = "micro",
    filter_expr: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Scores the extracted entities that may be overlapping or nested
    by looking in the spans returned by a given SpanGetter object.

    Parameters
    ----------
    examples: Examples
        The examples to score, either a tuple of (golds, preds) or a list of
        spacy.training.Example objects
    span_getter: SpanGetter
        The span getter to use to extract the spans from the document
    micro_key: str
        The key to use to store the micro-averaged results for spans of all types
    filter_expr: str
        The filter expression to use to filter the documents

    Returns
    -------
    Dict[str, Any]
    """
    examples = make_examples(examples)
    if filter_expr is not None:
        filter_fn = eval(f"lambda doc: {filter_expr}")
        examples = [eg for eg in examples if filter_fn(eg.reference)]
    labels = defaultdict(lambda: (set(), set()))
    labels["micro"] = (set(), set())
    for eg_idx, eg in enumerate(examples):
        for span in (
            span_getter(eg.predicted)
            if callable(span_getter)
            else get_spans(eg.predicted, span_getter)
        ):
            labels[span.label_][0].add((eg_idx, span.start, span.end, span.label_))
            labels[micro_key][0].add((eg_idx, span.start, span.end, span.label_))

        for span in (
            span_getter(eg.reference)
            if callable(span_getter)
            else get_spans(eg.reference, span_getter)
        ):
            labels[span.label_][1].add((eg_idx, span.start, span.end, span.label_))
            labels[micro_key][1].add((eg_idx, span.start, span.end, span.label_))

    return {name: prf(pred, gold) for name, (pred, gold) in labels.items()}


def ner_token_scorer(
    examples: Examples,
    span_getter: SpanGetter,
    micro_key: str = "micro",
    filter_expr: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Scores the extracted entities that may be overlapping or nested
    by looking in `doc.ents`, and `doc.spans`, and comparing the predicted
    and gold entities at the TOKEN level.

    Parameters
    ----------
    examples: Examples
        The examples to score, either a tuple of (golds, preds) or a list of
        spacy.training.Example objects
    span_getter: SpanGetter
        The span getter to use to extract the spans from the document
    micro_key: str
        The key to use to store the micro-averaged results for spans of all types
    filter_expr: str
        The filter expression to use to filter the documents

    Returns
    -------
    Dict[str, Any]
    """
    examples = make_examples(examples)
    if filter_expr is not None:
        filter_fn = eval(f"lambda doc: {filter_expr}")
        examples = [eg for eg in examples if filter_fn(eg.reference)]
    # label -> pred, gold
    labels = defaultdict(lambda: (set(), set()))
    labels["micro"] = (set(), set())
    for eg_idx, eg in enumerate(examples):
        for span in (
            span_getter(eg.predicted)
            if callable(span_getter)
            else get_spans(eg.predicted, span_getter)
        ):
            for i in range(span.start, span.end):
                labels[span.label_][0].add((eg_idx, i, span.label_))
                labels[micro_key][0].add((eg_idx, i, span.label_))

        for span in (
            span_getter(eg.reference)
            if callable(span_getter)
            else get_spans(eg.reference, span_getter)
        ):
            for i in range(span.start, span.end):
                labels[span.label_][1].add((eg_idx, i, span.label_))
                labels[micro_key][1].add((eg_idx, i, span.label_))

    return {name: prf(pred, gold) for name, (pred, gold) in labels.items()}


def dice(span1, span2):
    """
    Compute the Dice coefficient between two spans
    """
    intersection = max(0, min(span1.end, span2.end) - max(span1.start, span2.start))
    return 2 * intersection / (span1.end - span1.start + span2.end - span2.start)


def ner_overlap_scorer(
    examples: Examples,
    span_getter: SpanGetter,
    micro_key: str = "micro",
    filter_expr: Optional[str] = None,
    threshold: float = 0.5,
) -> Dict[str, Any]:
    """
    Scores the extracted entities that may be overlapping or nested
    by looking in `doc.ents`, and `doc.spans`, and comparing the predicted
    and gold entities and counting true when a predicted entity overlaps
    with a gold entity of the same label

    Parameters
    ----------
    examples: Examples
        The examples to score, either a tuple of (golds, preds) or a list of
        spacy.training.Example objects
    span_getter: SpanGetter
        The span getter to use to extract the spans from the document
    micro_key: str
        The key to use to store the micro-averaged results for spans of all types
    filter_expr: str
        The filter expression to use to filter the documents
    threshold: float
        The threshold to use to consider that two spans overlap

    Returns
    -------
    Dict[str, Any]
    """
    examples = make_examples(*examples)
    if filter_expr is not None:
        filter_fn = eval(f"lambda doc: {filter_expr}")
        examples = [eg for eg in examples if filter_fn(eg.reference)]
    # label -> pred, gold, matched_pred, matched_gold
    counters = defaultdict(lambda: [0, 0, 0, 0])
    counters["micro"] = [0, 0, 0, 0]
    for eg_idx, eg in enumerate(examples):
        pred_spans = set(
            span_getter(eg.predicted)
            if callable(span_getter)
            else get_spans(eg.predicted, span_getter)
        )
        gold_spans = set(
            span_getter(eg.reference)
            if callable(span_getter)
            else get_spans(eg.reference, span_getter)
        )
        for gold_span in gold_spans:
            counters[gold_span.label_][1] += 1
            counters[micro_key][1] += 1
        for pred_span in pred_spans:
            counters[pred_span.label_][0] += 1
            counters[micro_key][0] += 1

        for pred_span in pred_spans:
            overlaps = [
                (gold_span, dice(pred_span, gold_span))
                for gold_span in gold_spans
                if gold_span.label == pred_span.label
            ]
            matching_gold_span, overlap = max(
                overlaps, key=lambda x: x[1], default=(None, 0)
            )
            if matching_gold_span is not None and overlap >= threshold:
                counters[pred_span.label_][2] += 1
                counters[micro_key][2] += 1

        for gold_span in gold_spans:
            overlaps = [
                (pred_span, dice(gold_span, pred_span)) for pred_span in pred_spans
            ]
            matching_pred_span, overlap = max(
                overlaps, key=lambda x: x[1], default=(None, 0)
            )
            if (
                matching_pred_span is not None
                and overlap >= threshold
                and gold_span.label == matching_pred_span.label
            ):
                counters[gold_span.label_][3] += 1
                counters[micro_key][3] += 1

    results = {}
    for name, (num_pred, num_gold, true_pred, true_gold) in counters.items():
        p = (true_pred / num_pred) if num_pred else 1.0
        r = (true_gold / num_gold) if num_gold else 1.0
        f = (
            2 / (num_pred / true_pred + num_gold / true_gold)
            if true_pred and true_gold
            else 0.0
        )
        results[name] = {
            "f": f,
            "p": p,
            "r": r,
            "tp": true_pred,
            "support": num_gold,  # num gold
            "positives": num_pred,  # num predicted
        }
    return results


class NerScorer(abc.ABC):
    span_getter: SpanGetter

    def __call__(self, *examples) -> Dict[str, Any]:
        raise NotImplementedError()


@registry.scorers.register("eds.ner_exact_scorer")
class NerExactScorer(NerScorer):
    def __init__(
        self,
        span_getter: SpanGetterArg,
        micro_key: str = "micro",
        filter_expr: Optional[str] = None,
    ):
        self.span_getter = span_getter
        self.micro_key = micro_key
        self.filter_expr = filter_expr

    __init__.__doc__ = ner_exact_scorer.__doc__

    def __call__(self, *examples):
        return ner_exact_scorer(
            examples,
            span_getter=self.span_getter,
            micro_key=self.micro_key,
            filter_expr=self.filter_expr,
        )


@registry.scorers.register("eds.ner_token_scorer")
class NerTokenScorer(NerScorer):
    def __init__(
        self,
        span_getter: SpanGetterArg,
        micro_key: str = "micro",
        filter_expr: Optional[str] = None,
    ):
        self.span_getter = span_getter
        self.micro_key = micro_key
        self.filter_expr = filter_expr

    __init__.__doc__ = ner_token_scorer.__doc__

    def __call__(self, *examples):
        return ner_token_scorer(
            examples,
            span_getter=self.span_getter,
            micro_key=self.micro_key,
            filter_expr=self.filter_expr,
        )


@registry.scorers.register("eds.ner_overlap_scorer")
class NerOverlapScorer(NerScorer):
    def __init__(
        self,
        span_getter: SpanGetterArg,
        micro_key: str = "micro",
        filter_expr: Optional[str] = None,
        threshold: float = 0.5,
    ):
        self.span_getter = span_getter
        self.micro_key = micro_key
        self.filter_expr = filter_expr
        self.threshold = threshold

    __init__.__doc__ = ner_overlap_scorer.__doc__

    def __call__(self, *examples):
        return ner_overlap_scorer(
            examples,
            span_getter=self.span_getter,
            micro_key=self.micro_key,
            filter_expr=self.filter_expr,
            threshold=self.threshold,
        )


# For backward compatibility
create_ner_exact_scorer = NerExactScorer
create_ner_token_scorer = NerTokenScorer
create_ner_overlap_scorer = NerOverlapScorer
